from collections.abc import Callable
from dataclasses import dataclass
from typing import NamedTuple

import numpy as np

from src.types import INPUT_VECTOR


class Interval(NamedTuple):
    """Interval determining a number range (both inclusive)."""

    min: float
    max: float

    def random_point(self, dimensions: int, rng: np.random.Generator | None = None) -> INPUT_VECTOR:
        """Generate an N-dimensional random point, where all numbers lie within this interval.

        Args:
            dimensions: The amount of dimensions this point should have.
            rng: Random generator instance (None for a new rng).
        """
        if rng is None:
            rng = np.random.default_rng()

        return rng.uniform(self.min, self.max, size=dimensions)


@dataclass
class Function:
    """Class representing an N-dimensional function.

    This function can work with arbitrary number of dimensions, where each input dimension is restricted
    within the specified definition interval.
    """

    function: Callable[[INPUT_VECTOR], float]
    definition_interval: Interval

    def __call__(self, params: INPUT_VECTOR) -> float:
        """Evaluate the function."""
        return self.function(params)
