from collections.abc import Iterator

import numpy as np

from src.function import Function
from src.types import INPUT_VECTOR
from src.utils import generate_bounded_points


def min_search(
    function: Function,
    x0: INPUT_VECTOR,
    *,
    iterations: int,
    neighbors_count: int,
    std_dev: float,
    include_center: bool = True,
    rng: np.random.Generator | None = None,
) -> Iterator[INPUT_VECTOR]:
    """Search for the minimum value of the function using the Local Search / Stochastic Hill Climber algorithm.

    On each iteration, N neighboring points will be generated around the starting point. These points
    will then be evaluated on the function, finding the smallest one. This smallest point will become
    the new starting point, repeating until we run out of iterations.

    Params:
        x0: Starting point (N-dimensional).
        iterations: Maximum number of iterations.
        include_origin: When searching for the next minimum, should the origin point be checked too?
        neighbors_count: The amount of neighbor points.
        std_dev: Standard deviation for the normal distribution for neighbor generating.
        include_center:
            - When False, this algorithm works as Stochastic Hill Climber, always picking the minimum
              from the newly generated points only, ignoring the current point.
            - When True, this algorithm works as Local Search, picking the minimum from all of the
              newly generated points and also from the center point.
        rng: Random generator instance (None for a new rng).

    Yields:
        Minimum input vector (x) found so far, yielded from each iteration.
    """
    if rng is None:
        rng = np.random.default_rng()
    x_center = x0
    y_min = function(x_center) if include_center else float("inf")

    best_x = x_center
    best_y = y_min

    for _ in range(iterations):
        for point in generate_bounded_points(x_center, neighbors_count, std_dev, function.definition_interval, rng):
            y_point = function(point)
            if y_point < y_min:
                y_min = y_point
                x_center = point

        if y_min < best_y:
            best_y = y_min
            best_x = x_center

        # If we're using Stochastic Hill Climber, reset the minimum value, to only
        # pick it from the newly generated points, not the current center
        if not include_center:
            y_min = float("inf")

        yield best_x
