from typing import TypedDict
import hashlib

import matplotlib.pyplot as plt
import numpy as np
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QButtonGroup,
    QComboBox,
    QLabel,
    QLineEdit,
    QPushButton,
    QRadioButton,
    QSlider,
    QVBoxLayout,
    QWidget,
)

from src.function import Function
from src.functions import available_functions
from src.search import min_search
from src.types import INPUT_VECTOR


class OptimizationSettings(TypedDict):
    """Dictionary that holds the settings chosen for the optimization algorithm."""

    function: Function
    x0: INPUT_VECTOR
    iterations: int
    neighbors_count: int
    std_dev: float
    include_center: bool
    rng: np.random.Generator | None


class OptimizationUI(QWidget):
    """Class that represents the UI from which the user can control the optimization parameters."""

    def __init__(self) -> None:
        super().__init__()
        self.setWindowTitle("Optimization Algorithm Settings")
        self.setGeometry(100, 100, 400, 300)

        layout = QVBoxLayout()

        # Function Selection Dropdown
        self.function_label = QLabel("Select Function:")
        self.function_dropdown = QComboBox()
        self.function_dropdown.addItems(available_functions.keys())

        # Iterations Slider
        self.iter_label = QLabel("Iterations: 20")
        self.iter_slider = QSlider(Qt.Orientation.Horizontal)
        self.iter_slider.setRange(1, 1000)
        self.iter_slider.setValue(20)
        _ = self.iter_slider.valueChanged.connect(
            lambda: self.iter_label.setText(f"Iterations: {self.iter_slider.value()}"),
        )

        # StdDev Slider
        self.stddev_label = QLabel("StdDev: 3.0")
        self.stddev_slider = QSlider(Qt.Orientation.Horizontal)
        self.stddev_slider.setRange(0, 300)  # Scale by 10 to allow float precision
        self.stddev_slider.setValue(30)
        _ = self.stddev_slider.valueChanged.connect(
            lambda: self.stddev_label.setText(f"StdDev: {self.stddev_slider.value() / 10:.1f}"),
        )

        # Neighbors Count
        self.neighbors_label = QLabel("Neighbors: 3")
        self.neighbors_slider = QSlider(Qt.Orientation.Horizontal)
        self.neighbors_slider.setRange(1, 100)
        self.neighbors_slider.setValue(3)
        _ = self.neighbors_slider.valueChanged.connect(
            lambda: self.neighbors_label.setText(f"Neighbors: {self.neighbors_slider.value()}"),
        )

        # Dimensions Count
        self.dimensions_label = QLabel("Dimensions: 2")
        self.dimensions_slider = QSlider(Qt.Orientation.Horizontal)
        self.dimensions_slider.setRange(2, 100)
        _ = self.dimensions_slider.valueChanged.connect(
            lambda: self.dimensions_label.setText(f"Dimensions: {self.dimensions_slider.value()}"),
        )

        # Algorithm Selection (Radio Buttons)
        self.radio_group = QButtonGroup(self)
        self.local_search_radio = QRadioButton("Local Search (Include center points)")
        self.hill_climb_radio = QRadioButton("Stochastic Hill Climber (Only new points)")
        self.radio_group.addButton(self.local_search_radio)
        self.radio_group.addButton(self.hill_climb_radio)
        self.local_search_radio.setChecked(True)

        # RNG Seed Input
        self.seed_label = QLabel("RNG Seed:")
        self.seed_input = QLineEdit()
        self.seed_input.setPlaceholderText("Enter seed (optional)")

        # Run Button
        self.run_button = QPushButton("Run")
        _ = self.run_button.clicked.connect(self.run_algorithm)

        # Adding widgets to layout
        layout.addWidget(self.function_label)
        layout.addWidget(self.function_dropdown)
        layout.addWidget(self.iter_label)
        layout.addWidget(self.iter_slider)
        layout.addWidget(self.stddev_label)
        layout.addWidget(self.stddev_slider)
        layout.addWidget(self.neighbors_label)
        layout.addWidget(self.neighbors_slider)
        layout.addWidget(self.dimensions_label)
        layout.addWidget(self.dimensions_slider)
        layout.addWidget(self.local_search_radio)
        layout.addWidget(self.hill_climb_radio)
        layout.addWidget(self.seed_label)
        layout.addWidget(self.seed_input)
        layout.addWidget(self.run_button)

        self.setLayout(layout)

    @property
    def current_settings(self) -> OptimizationSettings:
        """Get currently selected settings for the optimization algorithm."""
        func_name = self.function_dropdown.currentText()
        function = available_functions[func_name]

        seed = self.seed_input.text()
        if seed != "":
            seed_int = int.from_bytes(hashlib.sha256(seed.encode("utf-8")).digest())
            bit_gen = np.random.PCG64(seed_int)
            rng = np.random.Generator(bit_gen)
        else:
            rng = None

        dims = self.dimensions_slider.value()
        x0 = function.definition_interval.random_point(dims, rng=rng)

        return OptimizationSettings(
            {
                "function": function,
                "x0": x0,
                "iterations": self.iter_slider.value(),
                "std_dev": self.stddev_slider.value() / 10,
                "neighbors_count": self.neighbors_slider.value(),
                "include_center": self.local_search_radio.isChecked(),
                "rng": rng,
            },
        )

    def run_algorithm(self) -> None:
        """Run the search algorithm according to selected settings.

        This is a call-back for pressing the run button.
        """
        settings = self.current_settings
        print(self.function_dropdown.currentText())

        values: list[float] = []
        x = None
        for x in min_search(**settings):
            y = settings["function"](x)
            values.append(y)

        if x is None:
            raise RuntimeError("Number of iterations can't be 0")

        self.plot_results(values, x)

    def plot_results(self, values: list[float], optimal_x: INPUT_VECTOR) -> None:
        """Plot the decreasing values using Matplotlib."""
        func_name = self.function_dropdown.currentText()

        _fig, ax = plt.subplots(figsize=(8, 5))
        _ = ax.plot(values, marker="o", linestyle="-", color="b", label="Function Value")
        _ = ax.set_xlabel("Iteration")
        _ = ax.set_ylabel("Minimum Value")
        _ = ax.set_title(f"Optimization Progress ({func_name})")
        _ = ax.legend()
        ax.grid(True)

        # Convert optimal_x to a readable string
        max_display_dim = 5  # Truncate if X has too many components (a lot of dimensions)
        if optimal_x.size > max_display_dim:
            optimal_x_str = np.array2string(optimal_x[:max_display_dim], precision=4, separator=", ")
            optimal_x_str = f"{optimal_x_str[:-1]}, ...]"  # Add ellipsis at the end
        else:
            optimal_x_str = np.array2string(optimal_x, precision=4, separator=", ")

        min_value = values[-1]

        table_data = [
            ["Min Value", f"{min_value:.6f}"],
            ["Optimal X", optimal_x_str],
        ]

        # Add table
        _ = plt.table(
            cellText=table_data,
            cellLoc="center",
            loc="bottom",
            bbox=[0.1, -0.4, 0.8, 0.2],  # Position table below plot
        )

        plt.subplots_adjust(bottom=0.35)  # Adjust layout to make room for the table
        plt.show()
