import numpy as np

from src.function import Interval
from src.types import INPUT_VECTOR


def generate_bounded_points(
    center: INPUT_VECTOR,
    num_points: int,
    std_dev: float,
    interval: Interval,
    rng: np.random.Generator | None = None,
) -> list[INPUT_VECTOR]:
    """Generate `num_points` random points around `center`, following a normal distribution.

    Args:
        center: The N-dimensional center point.
        num_points: Number of points to generate.
        std_dev: Standard deviation for the normal distribution.
        interval: Interval, determining the min/max values for each dimension.
        rng: Random generator instance (None for a new rng)

    Returns:
        A list of generated points.
    """
    dimensions = center.shape[0]
    valid_points: list[INPUT_VECTOR] = []

    if rng is None:
        rng = np.random.default_rng()

    while len(valid_points) < num_points:
        # Generate a batch of points (larger batch to reduce iterations)
        batch_size = (num_points - len(valid_points)) * 2  # Oversampling to reduce loop iterations
        candidates = rng.normal(loc=center, scale=std_dev, size=(batch_size, dimensions))

        # Keep only those within bounds
        mask = (candidates >= interval.min) & (candidates <= interval.max)
        valid_candidates = candidates[np.all(mask, axis=1)]

        # Add valid points to the list
        valid_points.extend(valid_candidates[: num_points - len(valid_points)])

    return valid_points
